Replaces hardcoded thresholds AND clamp bounds across 12 RL controllers
with observed-signal-driven ISV-slot bounds. Eliminates the architectural
failure mode that surfaced in walk-forward fold 0 (alpha-rl-m9cx5) and
fold 1 (alpha-rl-jgdh6): under Phase 4.5 advantage normalization, eight
controllers (PPO clip, target_tau, rollout_steps, entropy_coef, per_α,
gamma, reward_scale, q_distill_lambda) saturated at extrema within ~50
steps and stayed pinned for the rest of training — same pattern across
two different data slices, confirming the bug is structural rather than
data-size-dependent.
Mechanism: each controller's noise floor was hardcoded as a small
fraction of its target (`_NOISE_FLOOR_FRAC = 0.01f`), calibrated against
a pre-Phase-4.5 signal regime. Phase 4.5 normalization reduces operating
KL by ~50× — observed signal stays below the 1%-of-target floor, the
Schulman widen path fires continuously (asymmetric in the wrong
direction), and ε hits MAX 0.50 within ~50 training steps. ksll2's full-
data run (n_folds=1) happened to escape via a single above-band KL
observation that triggered tighten; both walk-forward folds (3 and 6
files) did not.
Per `feedback_adaptive_not_tuned`, `feedback_isv_for_adaptive_bounds`,
`pearl_controller_anchors_isv_driven`: every threshold now derives from
observed signal statistics (Welford online variance) rather than
constants calibrated against a prior signal regime. User explicitly
expanded scope mid-implementation: "if all clamps ISV bound should be
added to this spec and tasks" — applying the principle consistently
means hardcoded MIN/MAX clamp bounds count too, not just the saturating
noise floors. 12 controllers, single atomic commit per
`feedback_no_partial_refactor`.
Spec: docs/superpowers/specs/2026-05-30-adaptive-controller-floor-design.md
Plan: docs/superpowers/plans/2026-05-30-adaptive-controller-floor-plan.md
# New kernel
- rl_signal_variance_update.cu (80 LOC) — Welford online variance.
Single-thread single-block. Sentinel-zero skip per pearl. Per-controller
Welford triple (count, mean, M²) drives every adaptive noise floor.
# Per-controller refactors (12 .cu files)
- ppo_clip + target_tau + rollout_steps + entropy_coef + per_α —
adaptive noise floor = max(target × 0.5, sqrt(observed_var) × 2);
asymmetric Schulman (tighten on single observation, widen requires 3
consecutive below-band).
- gamma (Special G) — hardcoded GAMMA_MIN = 0.995 → adaptive via
Welford MEAN of trade duration. Per spec Q6 resolution: the Welford
mean's natural N-smoothing lag breaks the gamma↔trade_duration
feedback loop without explicit step-period gating. 100-observation
warmup falls back to EMA before Welford has enough samples.
- reward_scale (Special R) — asymmetric DECREASE rate cap (5% per
step) on both bootstrap-replace and Wiener-blend paths; bootstrap-
fraction floor (10% of bootstrap) until 100 trades close.
- q_distill_lambda (Special Q) — hardcoded MIN_LAMBDA = 0.05 → adaptive
via Welford on q_distill_kl_ema: max(0.001, std × 0.05).
- v_blend_alpha (Phase 4.4) — 5 hardcoded constants → 5 ISV slots
(DEAD_SIGNAL_FLOOR, TARGET_TRACK_RATIO, SCHULMAN_STEP, EMA_ALPHA,
BOOTSTRAP_ALPHA) + adaptive dead-signal floor from V_scalar magnitude
variance.
- ppo_ratio_clamp — adaptive MIN/MAX from observed log-ratio variance.
Architectural 2.0 absolute floor preserved (don't degenerate to
vanilla policy gradient).
- reward_clamp — V_BOUND_FLOOR, V_BOUND_EWMA_ALPHA, MIN_WIN, MIN_RATIO,
MAX_RATIO, MIN/MAX_MARGIN, MARGIN_TOLERANCE/ADJUST_RATE,
CLIP_RATE_EMA_ALPHA → 10 ISV slots.
- gate_threshold — hardcoded alpha = 0.01 → ISV slot 638.
- Clamp-bound expansion: EPS_MIN/MAX, TAU_MIN, ROLLOUT_MIN/MAX,
COEF_MIN/MAX, PER_ALPHA_MIN/MAX, GAMMA_MAX, MAX_LAMBDA, KL_TOLERANCE,
LAMBDA_RAMP_RATE, LAMBDA_DECAY_RATE → all ISV.
- WIENER_ALPHA_FLOOR shared across 9 controllers → single ISV slot 659.
# Trainer integration (integrated.rs)
- rl_signal_variance_update kernel loaded + helper method
`launch_rl_signal_variance_update`.
- Per-step Welford launches for 9 controller inputs, placed between
EMA producers and rl_fused_controllers in the per-step pipeline.
- Input-slot lookup array for rl_fused_controllers updated: rollout_steps
now consumes RL_ADV_VAR_PRE_NORM_INDEX (emitted by
rl_advantage_normalize before in-place normalize) instead of the
post-norm advantage_var_ratio (definitionally ~0 under Phase 4.5).
- ~30 new ISV bootstrap entries in with_controllers_bootstrapped.
# ISV slot allocation (isv_slots.rs)
- 72 new slots, RL_SLOTS_END 588 → 660. +288 bytes mapped-pinned.
- 9 Welford variance triples + 5 asymmetric Schulman counters +
3 new input signals (var_pre_norm, gamma_min_adaptive, reward_mag) +
v_blend (5 + 3 Welford) + ppo_ratio_clamp (1 + 3 Welford) +
reward_clamp (5) + gate_threshold (1) + q_distill (1) + 20 clamp bounds +
shared wiener floor.
# Bootstrap-clamp consistency fix (caught by Task 16 testing)
The original draft bootstrapped RL_EPS_BOOTSTRAP at 0.01 (= KL target),
but EPS_MIN was 0.05 — bootstrap value below clamp range. First post-
bootstrap step always snapped ε to MIN regardless of signal direction
("snap to MIN" behavior the unit tests surfaced). Fixed by bootstrapping
ε at MIN (0.05) so asymmetric Schulman operates from a valid state.
# Validation
- cargo build --release: clean
- cargo build --tests: clean
- 3 GPU Welford kernel tests (G1 constant→0var, G2 sequence→known var,
G3 sentinel skip): all pass
- 5 GPU adaptive-floor invariant tests (G3 PPO clip holds at bootstrap
when signal below floor, G4 tightens on single above-band, G5 widens
only after 3 consecutive below-band, G6 post-warmup uses Welford mean,
G6 pre-warmup uses EMA): all pass
- integrated_trainer_smoke (full GPU pipeline): passes
- 1k local smoke at b=128: 1000/1000 steps, completed_clean, no NaN,
controllers genuinely adapting (γ 0.90→0.974, per_α 0.40→0.52,
q_distill_λ 0.05→0.21, ε held at 0.05 = MIN per Phase 4.5 small-KL
regime — was previously stuck at MAX 0.50 throughout fold 0/1)
- compute-sanitizer memcheck b=128 5 steps: 0 errors
Cluster validation (G7-G9) submitted as follow-up runs.
🤖 Generated with [Claude Code](https://claude.com/claude-code)
Co-Authored-By: Claude Opus 4.7 <noreply@anthropic.com>
89 lines
4.1 KiB
Plaintext
89 lines
4.1 KiB
Plaintext
// rl_q_distill_lambda_controller.cu — adaptive λ_distill controller.
|
||
//
|
||
// Adapts the Q→π distillation strength (λ, slot 486) from the KL
|
||
// divergence between Q's Boltzmann policy and π's softmax (slot 488).
|
||
//
|
||
// Controller logic (asymmetric bounded step):
|
||
//
|
||
// if KL > target × TOLERANCE → λ *= RAMP_RATE (Q signal not
|
||
// landing, pull harder)
|
||
// if KL < target / TOLERANCE → λ *= DECAY_RATE (aligned, ease off)
|
||
// else → λ unchanged (dead-zone)
|
||
//
|
||
// CRITICAL DESIGN: decay is 100× slower than ramp. This prevents the
|
||
// self-defeating oscillation where λ ramps up → Q and π align → λ
|
||
// collapses → they diverge again. The asymmetry means coupling
|
||
// establishes quickly but persists for thousands of steps, giving
|
||
// Q time to teach π meaningful action preferences.
|
||
//
|
||
// Per `feedback_no_atomicadd`: single-thread kernel, no atomics.
|
||
// Per `feedback_cpu_is_read_only`: all state in ISV.
|
||
|
||
#define RL_Q_DISTILL_LAMBDA_INDEX 486
|
||
#define RL_Q_DISTILL_KL_EMA_INDEX 488
|
||
#define RL_Q_DISTILL_KL_TARGET_INDEX 491
|
||
|
||
// Adaptive controller floors (spec 2026-05-30 Special case Q):
|
||
// hardcoded `MIN_LAMBDA = 0.05f` floor replaced with adaptive bound
|
||
// derived from Welford variance on q_distill_kl_ema. Phase 4.5 reduces
|
||
// KL signal magnitudes which this controller consumes; the "below target"
|
||
// path fires continuously, decaying λ to MIN. Adaptive bound
|
||
// `max(0.001, sqrt(var) × 0.05)` lets λ decay to a level proportional to
|
||
// the actual signal noise rather than the hardcoded 0.05 floor.
|
||
#define RL_Q_DISTILL_KL_VAR_COUNT_INDEX 618
|
||
#define RL_Q_DISTILL_KL_VAR_M2_INDEX 620
|
||
#define RL_Q_DISTILL_LAMBDA_MIN_ADAPTIVE_INDEX 639
|
||
|
||
// MAX_LAMBDA, KL_TOLERANCE, LAMBDA_RAMP_RATE, LAMBDA_DECAY_RATE are now
|
||
// ISV-driven per the 2026-05-30 clamp-bound extension. ADAPTIVE_MIN_*
|
||
// remain hardcoded — they're new constants from the noise-floor design
|
||
// itself per spec exemption ("not clamp bounds, new pattern constants").
|
||
#define RL_Q_DISTILL_LAMBDA_MAX_INDEX 650
|
||
#define RL_Q_DISTILL_KL_TOLERANCE_INDEX 651
|
||
#define RL_Q_DISTILL_LAMBDA_RAMP_RATE_INDEX 652
|
||
#define RL_Q_DISTILL_LAMBDA_DECAY_RATE_INDEX 653
|
||
#define ADAPTIVE_MIN_ABSOLUTE 0.001f
|
||
#define ADAPTIVE_MIN_STD_SCALE 0.05f
|
||
|
||
extern "C" __global__ void rl_q_distill_lambda_controller(
|
||
float* __restrict__ isv
|
||
) {
|
||
if (threadIdx.x != 0 || blockIdx.x != 0) return;
|
||
|
||
const float kl_observed = isv[RL_Q_DISTILL_KL_EMA_INDEX];
|
||
const float kl_target = isv[RL_Q_DISTILL_KL_TARGET_INDEX];
|
||
float lambda = isv[RL_Q_DISTILL_LAMBDA_INDEX];
|
||
|
||
if (kl_observed <= 0.0f || kl_target <= 0.0f || lambda <= 0.0f) return;
|
||
|
||
// Adaptive MIN bound (spec 2026-05-30 Special case Q): replaces
|
||
// hardcoded `MIN_LAMBDA = 0.05f`. Welford sample variance =
|
||
// M² / (count − 1) when count > 1. Adaptive min scales with signal
|
||
// std so λ can decay proportional to actual KL noise rather than
|
||
// pegging at an arbitrary 0.05 floor.
|
||
const float kl_var_count = isv[RL_Q_DISTILL_KL_VAR_COUNT_INDEX];
|
||
const float kl_var = (kl_var_count > 1.0f)
|
||
? isv[RL_Q_DISTILL_KL_VAR_M2_INDEX] / (kl_var_count - 1.0f)
|
||
: 0.0f;
|
||
const float kl_std = sqrtf(kl_var);
|
||
const float adaptive_min = fmaxf(ADAPTIVE_MIN_ABSOLUTE,
|
||
kl_std * ADAPTIVE_MIN_STD_SCALE);
|
||
isv[RL_Q_DISTILL_LAMBDA_MIN_ADAPTIVE_INDEX] = adaptive_min;
|
||
|
||
const float max_lambda = isv[RL_Q_DISTILL_LAMBDA_MAX_INDEX];
|
||
const float kl_tolerance = isv[RL_Q_DISTILL_KL_TOLERANCE_INDEX];
|
||
const float lambda_ramp_rate = isv[RL_Q_DISTILL_LAMBDA_RAMP_RATE_INDEX];
|
||
const float lambda_decay_rate = isv[RL_Q_DISTILL_LAMBDA_DECAY_RATE_INDEX];
|
||
|
||
const float upper = kl_target * kl_tolerance;
|
||
const float lower = kl_target / kl_tolerance;
|
||
|
||
if (kl_observed > upper) {
|
||
lambda = fminf(max_lambda, lambda * lambda_ramp_rate);
|
||
} else if (kl_observed < lower) {
|
||
lambda = fmaxf(adaptive_min, lambda * lambda_decay_rate);
|
||
}
|
||
|
||
isv[RL_Q_DISTILL_LAMBDA_INDEX] = lambda;
|
||
}
|